# pylint: skip-file
"""
///
/// @file MD10State.py
///
/// @brief AUTOGENERATED FILE! ONLY EDIT IF YOU ARE A MACHINE!
///
/// @core 0bad34709094958919bfe30e6c0ca13c7682c49a
///
/// @build df4ac791e9dd25df235f8ed979101d1d2e75b19b
///
/// @date 2021-12-16 14:53:15
///
/// @author Dephy, Inc.
"""
from ctypes import Structure, c_int


class MD10State(Structure):
    _pack_ = 1
    _fields_ = [
        ("rigid", c_int),
        ("id", c_int),
        ("state_time", c_int),
        ("accelx", c_int),
        ("accely", c_int),
        ("accelz", c_int),
        ("gyrox", c_int),
        ("gyroy", c_int),
        ("gyroz", c_int),
        ("mot_ang", c_int),
        ("mot_vel", c_int),
        ("mot_acc", c_int),
        ("mot_cur", c_int),
        ("mot_volt", c_int),
        ("batt_volt", c_int),
        ("batt_curr", c_int),
        ("temperature", c_int),
        ("status_mn", c_int),
        ("status_ex", c_int),
        ("status_re", c_int),
        ("genvar_0", c_int),
        ("genvar_1", c_int),
        ("genvar_2", c_int),
        ("genvar_3", c_int),
        ("genvar_4", c_int),
        ("genvar_5", c_int),
        ("genvar_6", c_int),
        ("genvar_7", c_int),
        ("genvar_8", c_int),
        ("genvar_9", c_int),
        ("genvar_10", c_int),
        ("genvar_11", c_int),
        ("genvar_12", c_int),
        ("genvar_13", c_int),
        ("genvar_14", c_int),
        ("ank_ang", c_int),
        ("ank_vel", c_int),
        ("shank_ang", c_int),
        ("shank_vel", c_int),
        ("global_shank_ang", c_int),
        ("mot_from_ank", c_int),
        ("ank_from_mot", c_int),
        ("trans_ratio", c_int),
        ("ank_torque", c_int),
        ("step_count", c_int),
        ("step_time", c_int),
        ("gait_state", c_int),
        ("intermediate_state", c_int),
        ("movement", c_int),
        ("SystemTime", c_int),
    ]



"""
Demo 3: Current Control

Here we loop over an array of currents, demonstrating how to:

    * Command the motor current
    * Command the motor to stop
    * Print device data
"""
import platform
import subprocess as sub
from time import sleep

import numpy as np

from flexsea.device import Device


# ============================================
#                   clear
# ============================================
def clear() -> None:
    try:
        sub.run(
            ["cls" if "windows" == platform.system().lower() else "clear"], check=True
        )
    except FileNotFoundError:
        sub.run(["clear"], check=True)


if "windows" == platform.system().lower():
    msg = "WARNING: these demos may not function properly on Windows "
    msg += "due to the way the OS handles timing. They are best run on "
    msg += "linux."
    print(msg)


# Instantiate and connect to device; see demo1.py
device = Device()
device.open()


# Start streaming; see demo2.py
frequency = 100
device.start_streaming(frequency)


# Set device gains; see demo2.py
gains = {
    "kp": 40,
    "ki": 400,
    "kd": 0,
    "k": 0,
    "b": 0,
    "ff": 0,
}
device.set_gains(**gains)


# Create the array of currents we want to loop over (in milliamps)
maxCurrent = 1000
nCurrents = 50
currents = np.linspace(0, maxCurrent, nCurrents)


# Ramp up over the desired currents, hold, then ramp back down
commandDelay = 0.1
holdTime = 3

for current in currents:
    # We set the current with the `command_motor_current` method, which
    # takes in the desired current in milliamps
    device.command_motor_current(int(current))
    sleep(commandDelay)
    # There are two ways to print device data: we can either call the
    # `print` method to print the most recent data, or we can call the
    # `print` method with a data dictionary passed as an argument in
    # order to print the data dictionary instead of the most recent
    # data
    device.print()
    clear()

sleep(holdTime)

# During the ramp down phase, we here allow for a keyboard interrupt in
# order to demonstrate how the motor can be manually stopped via the
# `stop_motor` method
try:
    for current in currents[-1::-1]:
        device.command_motor_current(int(current))
        sleep(commandDelay)
        # Here we show the second way of calling `print` in order to
        # easily print data that was read at some point in the past
        data = device.read()
        device.print(data)
        clear()
except KeyboardInterrupt:
    device.stop_motor()
    sleep(commandDelay)

device.close()



# pylint: skip-file
"""
///
/// @file BMSState.py
///
/// @brief AUTOGENERATED FILE! ONLY EDIT IF YOU ARE A MACHINE!
///
/// @core 0bad34709094958919bfe30e6c0ca13c7682c49a
///
/// @build df4ac791e9dd25df235f8ed979101d1d2e75b19b
///
/// @date 2020-08-19 10:28:50
///
/// @author Dephy, Inc.
"""
from ctypes import Structure, c_int


class BMSState(Structure):
    _pack_ = 1
    _fields_ = [
        ("bms_state_0", c_int),
        ("id", c_int),
        ("state_time", c_int),
        ("cells_0_mv", c_int),
        ("cells_1_mv", c_int),
        ("cells_2_mv", c_int),
        ("cells_3_mv", c_int),
        ("cells_4_mv", c_int),
        ("cells_5_mv", c_int),
        ("cells_6_mv", c_int),
        ("cells_7_mv", c_int),
        ("cells_8_mv", c_int),
        ("bms_state", c_int),
        ("error", c_int),
        ("current", c_int),
        ("stack_voltage", c_int),
        ("temperature_0", c_int),
        ("temperature_1", c_int),
        ("temperature_2", c_int),
        ("pack_imbalance", c_int),
        ("balancing", c_int),
        ("timer", c_int),
        ("user_interface", c_int),
        ("bq_sys_stat", c_int),
        ("bq_sys_ctrl1", c_int),
        ("bq_sys_ctrl2", c_int),
        ("fw_version", c_int),
        ("genvar_0_", c_int),
        ("genvar_1_", c_int),
        ("SystemTime", c_int),
    ]



import ctypes as c

from flexsea import firmware as fw
from flexsea.specs.device_specs import all_devices as fxd


read_functions = {
    "actpack": {
        "name": "fxReadDevice",
        "argTypes": [c.c_uint, c.POINTER(fxd.ActPackState)],
        "returnType": c.c_int,
        "all_name": "fxReadDeviceAll",
        "all_argTypes": [c.c_uint, c.POINTER(fxd.ActPackState), c.c_uint],
        "all_returnType": c.c_int,
    },
    "exo": {
        "name": "fxReadDevice",
        "argTypes": [c.c_uint, c.POINTER(fxd.EB60State)],
        "returnType": c.c_int,
        "all_name": "fxReadDeviceAll",
        "all_argTypes": [c.c_uint, c.POINTER(fxd.EB60State), c.c_uint],
        "all_returnType": c.c_int,
    },
}


v9_1_0 = {
    "commands": {
        "is_open": {
            "name": "fxIsOpen",
            "argTypes": [
                c.c_uint,
            ],
            "returnType": c.c_bool,
        },
        "is_streaming": {
            "name": "fxIsStreaming",
            "argTypes": [
                c.c_uint,
            ],
            "returnType": c.c_bool,
        },
        "open": {
            "name": "fxOpen",
            "argTypes": [c.c_char_p, c.c_uint, c.c_uint],
            "returnType": c.c_int,
        },
        "close": {
            "name": "fxClose",
            "argTypes": [
                c.c_uint,
            ],
            "returnType": c.c_int,
        },
        "start_streaming": {
            "name": "fxStartStreaming",
            "argTypes": [c.c_uint, c.c_uint, c.c_bool],
            "returnType": c.c_int,
        },
        "start_streaming_with_safety": {
            "name": "fxStartStreamingWithSafety",
            "argTypes": [c.c_uint, c.c_uint, c.c_bool, c.c_uint16],
            "returnType": c.c_int,
        },
        "stop_streaming": {
            "name": "fxStopStreaming",
            "argTypes": [
                c.c_uint,
            ],
            "returnType": c.c_int,
        },
        "set_gains": {
            "name": "fxSetGains",
            "argTypes": [
                c.c_uint,
                c.c_uint,
                c.c_uint,
                c.c_uint,
                c.c_uint,
                c.c_uint,
                c.c_uint,
            ],
            "returnType": c.c_int,
        },
        "send_motor_command": {
            "name": "fxSendMotorCommand",
            "argTypes": [c.c_uint, c.c_int, c.c_int],
            "returnType": c.c_int,
        },
        "find_poles": {
            "name": "fxFindPoles",
            "argTypes": [
                c.c_uint,
            ],
            "returnType": c.c_int,
        },
        "activate_bootloader": {
            "name": "fxActivateBootloader",
            "argTypes": [c.c_uint, c.c_uint8],
            "returnType": c.c_int,
        },
        "is_bootloader_activated": {
            "name": "fxIsBootloaderActivated",
            "argTypes": [
                c.c_uint,
            ],
            "returnType": c.c_int,
        },
        "request_firmware_version": {
            "name": "fxRequestFirmwareVersion",
            "argTypes": [
                c.c_uint,
            ],
            "returnType": c.c_int,
        },
        "get_last_received_firmware_version": {
            "name": "fxGetLastReceivedFirmwareVersion",
            "argTypes": [
                c.c_uint,
            ],
            "returnType": fw.Firmware,
        },
        "get_device_type_value": {
            "name": "fxGetAppType",
            "argTypes": [
                c.c_uint,
            ],
            "returnType": c.c_int,
        },
        "get_read_data_queue_size": {
            "name": "fxGetReadDataQueueSize",
            "argTypes": [
                c.c_uint,
            ],
            "returnType": c.c_int,
        },
        "set_read_data_queue_size": {
            "name": "fxSetReadDataQueueSize",
            "argTypes": [c.c_uint, c.c_uint],
            "returnType": c.c_uint,
        },
    },
    "read_functions": read_functions,
}



# pylint: skip-file
"""
///
/// @file NetMasterState.py
///
/// @brief AUTOGENERATED FILE! ONLY EDIT IF YOU ARE A MACHINE!
///
/// @core 0bad34709094958919bfe30e6c0ca13c7682c49a
///
/// @build df4ac791e9dd25df235f8ed979101d1d2e75b19b
///
/// @date 2020-06-22 14:42:45
///
/// @author Dephy, Inc.
"""
from ctypes import Structure, c_int


class NetMasterState(Structure):
    _pack_ = 1
    _fields_ = [
        ("netmaster", c_int),
        ("id", c_int),
        ("state_time", c_int),
        ("genvar_0", c_int),
        ("genvar_1", c_int),
        ("genvar_2", c_int),
        ("genvar_3", c_int),
        ("status", c_int),
        ("a_accelx", c_int),
        ("a_accely", c_int),
        ("a_accelz", c_int),
        ("a_gyrox", c_int),
        ("a_gyroy", c_int),
        ("a_gyroz", c_int),
        ("a_pressure", c_int),
        ("a_status", c_int),
        ("b_accelx", c_int),
        ("b_accely", c_int),
        ("b_accelz", c_int),
        ("b_gyrox", c_int),
        ("b_gyroy", c_int),
        ("b_gyroz", c_int),
        ("b_pressure", c_int),
        ("b_status", c_int),
        ("c_accelx", c_int),
        ("c_accely", c_int),
        ("c_accelz", c_int),
        ("c_gyrox", c_int),
        ("c_gyroy", c_int),
        ("c_gyroz", c_int),
        ("c_pressure", c_int),
        ("c_status", c_int),
        ("d_accelx", c_int),
        ("d_accely", c_int),
        ("d_accelz", c_int),
        ("d_gyrox", c_int),
        ("d_gyroy", c_int),
        ("d_gyroz", c_int),
        ("d_pressure", c_int),
        ("d_status", c_int),
        ("e_accelx", c_int),
        ("e_accely", c_int),
        ("e_accelz", c_int),
        ("e_gyrox", c_int),
        ("e_gyroy", c_int),
        ("e_gyroz", c_int),
        ("e_pressure", c_int),
        ("e_status", c_int),
        ("f_accelx", c_int),
        ("f_accely", c_int),
        ("f_accelz", c_int),
        ("f_gyrox", c_int),
        ("f_gyroy", c_int),
        ("f_gyroz", c_int),
        ("f_pressure", c_int),
        ("f_status", c_int),
        ("g_accelx", c_int),
        ("g_accely", c_int),
        ("g_accelz", c_int),
        ("g_gyrox", c_int),
        ("g_gyroy", c_int),
        ("g_gyroz", c_int),
        ("g_pressure", c_int),
        ("g_status", c_int),
        ("h_accelx", c_int),
        ("h_accely", c_int),
        ("h_accelz", c_int),
        ("h_gyrox", c_int),
        ("h_gyroy", c_int),
        ("h_gyroz", c_int),
        ("h_pressure", c_int),
        ("h_status", c_int),
        ("SystemTime", c_int),
    ]



"""
Demo 5: High Speed

Runs the motor at a high speed, demonstrating how to:

    * Control which version of the pre-compiled C library is used
"""
import platform
from time import sleep

import numpy as np
import pandas as pd
import seaborn as sns

from flexsea.device import Device


if "windows" == platform.system().lower():
    msg = "WARNING: these demos may not function properly on Windows "
    msg += "due to the way the OS handles timing. They are best run on "
    msg += "linux."
    print(msg)


# Instantiate and connect to the device; see demo1.py
# Here we see that we can change which version of the pre-compiled C
# library is used by passing in a semantic version string to the
# `cLibVersion` keyword argument
device = Device(cLibVersion="9.1.0")
device.open()


# Start streaming; see demo2.py
frequency = 100
device.start_streaming(frequency)


# Set gains; see demo2.py
gains = {
    "kp": 40,
    "ki": 400,
    "kd": 0,
    "k": 0,
    "b": 0,
    "ff": 128,
}
device.set_gains(**gains)


# Generate current values to command following a sine wave
waveAmplitude = 500
waveFrequency = 5
commandFrequency = 500

nSamples = int(commandFrequency / waveFrequency)
x = np.linspace(-np.pi, np.pi, nSamples)
currents = waveAmplitude * np.sin(x)


# Loop over the currents several times
nLoops = 4
commandDelay = 1 / commandFrequency
cycleDelay = 0.1

measuredCurrent = []
desiredCurrent = []
deviceTime = []

for _ in range(nLoops):
    for current in currents:
        sleep(commandDelay)

        data = device.read()

        measuredCurrent.append(data["mot_cur"])
        desiredCurrent.append(current)
        deviceTime.append(data["state_time"])

        device.command_motor_current(int(current))

        # Delay between cycles
        for __ in range(int(cycleDelay / commandDelay)):
            sleep(commandDelay)

            data = device.read()

            measuredCurrent.append(data["mot_cur"])
            desiredCurrent.append(current)
            deviceTime.append(data["state_time"])


# Plot; see demo2.py
nValues = len(deviceTime)
t = np.concatenate((deviceTime, deviceTime))
currentType = ["Desired Current"] * nValues + ["Measured Current"] * nValues
current = np.concatenate((desiredCurrent, measuredCurrent))

data = {
    "Time (ms)": t,
    "Current Type": currentType,
    "Current (mA)": current,
}

df = pd.DataFrame(data)

plot = sns.relplot(
    data=df, x="Time (ms)", y="Current (mA)", hue="Current Type", kind="line"
)

plot.figure.savefig("high_speed.png")

print("Data plot saved as: 'high_speed.png'")



"""
Demo 1: Open control

Here we loop over an array of voltages, demonstrating how to:

    * Instantiate the Device class
    * Establish a connection between the device and the computer
    * Command the motor voltage
    * Clean up once we're done
"""
import platform
from time import sleep

import numpy as np

from flexsea.device import Device


if "windows" == platform.system().lower():
    msg = "WARNING: these demos may not function properly on Windows "
    msg += "due to the way the OS handles timing. They are best run on "
    msg += "linux."
    print(msg)


# We begin by creating an instance of the Device class. By default,
# flexsea will attempt to find the connected device for you. It will
# also use a default value for the baud rate and version of the pre-
# compiled C library, both of which are set in flexsea.config. You
# can override these values by setting the appropriate keyword
# arguments in the constructor. If you have more than one device
# connected, you should manually specify the port for each
device = Device()


# Once instantiated, we have to establish a connection between the
# computer and the device
device.open()


# Generate the array of voltages (millivolts) to loop over
maxVoltage = 3000
nVoltages = 10
voltages = np.linspace(0, maxVoltage, nVoltages)


# Ramp up and down the voltages several times
nCycles = 5

for _ in range(nCycles):
    # Ramp up
    for voltage in voltages:
        # We use the command_motor_voltage method to set the voltage
        device.command_motor_voltage(int(voltage))
        # We now perform a brief sleep in order to give the device time
        # to process and execute our command
        sleep(0.1)

    # Ramp down
    for voltage in voltages[-1::-1]:
        device.command_motor_voltage(int(voltage))
        sleep(0.1)


# Once we're done using the device, we should clean up by calling the
# close method
device.close()



"""
Demo 4: Impedance Control

Oscillates between two different positions by controlling the motor's
impedance, demonstrating how to:

    * Command motor impedance
"""
import platform
from time import sleep

import numpy as np
import pandas as pd
import seaborn as sns

from flexsea.device import Device


if "windows" == platform.system().lower():
    msg = "WARNING: these demos may not function properly on Windows "
    msg += "due to the way the OS handles timing. They are best run on "
    msg += "linux."
    print(msg)


# Instantiate and connect to the device; see demo1.py
device = Device()
device.open()


# Start streaming; see demo2.py
frequency = 100
device.start_streaming(frequency)


# Read from the device; see demo2.py
data = device.read()


# Set positions
offset = 7500
pos0 = data["mot_ang"]
positions = [pos0, pos0 + offset]


# Set gains; see demo2.py
gains = {
    "kp": 40,
    "ki": 400,
    "kd": 0,
    "k": 600,
    "b": 300,
    "ff": 128,
}
device.set_gains(**gains)


# Move between the two positions; see demo2.py
runTime = 10
transitionTime = 0.8
commandDelay = 0.02
positionIndex = 0
bGainDelta = 150

measuredPosition = []
desiredPosition = []
deviceTime = []

nLoops = int(runTime / commandDelay)
transitionSteps = int(transitionTime / commandDelay)

for i in range(nLoops):
    data = device.read()

    measuredPosition.append(data["mot_ang"])
    deviceTime.append(data["state_time"])
    desiredPosition.append(positions[positionIndex])

    if i % transitionSteps == 0:
        gains["b"] += bGainDelta
        device.set_gains(**gains)
        positionIndex = (positionIndex + 1) % len(positions)
        # Here we command the motor impedance via the
        # `command_motor_impedance` method, which takes in a motor
        # position in ticks
        device.command_motor_impedance(int(positions[positionIndex]))

    sleep(commandDelay)

device.close()


# Plot; see demo2.py
nValues = len(deviceTime)
t = np.concatenate((deviceTime, deviceTime))
positionType = ["Desired Position"] * nValues + ["Measured Position"] * nValues
position = np.concatenate((desiredPosition, measuredPosition))

data = {
    "Time (ms)": t,
    "Position Type": positionType,
    "Motor Position (ticks)": position,
}

df = pd.DataFrame(data)

plot = sns.relplot(
    data=df, x="Time (ms)", y="Motor Position (ticks)", hue="Position Type", kind="line"
)

plot.figure.savefig("impedance_demo.png")

print("Data plot saved as: 'impedance_demo.png'")



# pylint: skip-file
"""
///
/// @file OPORPState.py
///
/// @brief AUTOGENERATED FILE! ONLY EDIT IF YOU ARE A MACHINE!
///
/// @core 31f85f4604524f5e70b1d8605021b0952ca8b26b
///
/// @build fb8c2409d776759b5191b6bacaabe51a49ade1cf
///
/// @date 2021-05-10 10:03:07
///
/// @author Dephy, Inc.
"""
from ctypes import Structure, c_int


class OPORPState(Structure):
    _pack_ = 1
    _fields_ = [
        ("rigid", c_int),
        ("id", c_int),
        ("state_time", c_int),
        ("accelx", c_int),
        ("accely", c_int),
        ("accelz", c_int),
        ("gyrox", c_int),
        ("gyroy", c_int),
        ("gyroz", c_int),
        ("mot_ang", c_int),
        ("mot_vel", c_int),
        ("mot_acc", c_int),
        ("mot_cur", c_int),
        ("mot_volt", c_int),
        ("batt_volt", c_int),
        ("batt_curr", c_int),
        ("temperature", c_int),
        ("status_mn", c_int),
        ("status_ex", c_int),
        ("status_re", c_int),
        ("genvar_0", c_int),
        ("genvar_1", c_int),
        ("genvar_2", c_int),
        ("genvar_3", c_int),
        ("genvar_4", c_int),
        ("genvar_5", c_int),
        ("genvar_6", c_int),
        ("genvar_7", c_int),
        ("genvar_8", c_int),
        ("genvar_9", c_int),
        ("ank_ang", c_int),
        ("ank_vel", c_int),
        ("SystemTime", c_int),
    ]



import ctypes as c
from typing import List

from . import enums as fxe
from .dephy_device import DephyDevice
from .specs.api_spec import apiSpec


# ============================================
#                LegacyDevice
# ============================================
class LegacyDevice(DephyDevice):
    """
    Representation of one of Dephy's devices using old comms and
    functionality (e.g., device doesn't know side info).
    """

    SUCCESS = fxe.legacyDeviceErrorCodes["SUCCESS"]
    FAILURE = fxe.legacyDeviceErrorCodes["FAILURE"]
    INVALID_PARAM = fxe.legacyDeviceErrorCodes["INVALID_PARAM"]
    INVALID_DEVICE = fxe.legacyDeviceErrorCodes["INVALID_DEVICE"]
    NOT_STREAMING = fxe.legacyDeviceErrorCodes["NOT_STREAMING"]

    # -----
    # constructor
    # -----
    def __init__(
        self,
        port: str,
        baudRate: int,
        cLibVersion: str,
        logLevel: int,
        loggingEnabled: bool,
        libFile: str = "",
    ) -> None:

        super().__init__(port, baudRate, cLibVersion, logLevel, loggingEnabled, libFile)
        self._state: c.Structure | None = None

    # -----
    # _setup
    # -----
    def _setup(self) -> None:
        """
        Gets device meta-info after establishing a connection. This
        is the part of opening that differs from that of Device.
        """
        self._deviceName = self.deviceName

        if self._deviceName in fxe.hasHabsLegacy:
            self.hasHabs = True

        self._state = fxe.deviceStateDicts[self._deviceName]()

        # The read function is set here and not with other c functions
        # because we need the device name, which we can't get without
        # calling open, for which we need the c library loaded
        rf = apiSpec[self.cLibVersion]["read_functions"][self._deviceName]

        for key in ("", "all_"):
            func = getattr(self._clib, rf[key + "name"])
            func.argtypes = rf[key + "argTypes"]
            func.restype = rf[key + "returnType"]
            setattr(self._clib, key + "read", func)

    # -----
    # deviceName
    # -----
    @property
    def deviceName(self) -> str:
        if self._deviceName:
            return self._deviceName

        deviceTypeValue = self._clib.get_device_type_value(self.deviceId)
        return fxe.deviceNames[deviceTypeValue]

    # -----
    # deviceSide
    # -----
    @property
    def deviceSide(self) -> str:
        raise RuntimeError("Legacy devices don't have this functionality.")

    # -----
    # libs_version
    # -----
    @property
    def libsVersion(self) -> str:
        raise RuntimeError("Legacy devices don't have this functionality.")

    # -----
    # _read
    # -----
    def _read(self) -> dict:
        if self._clib.read(self.deviceId, c.byref(self._state)) != self.SUCCESS.value:
            raise RuntimeError("Error: read command failed.")
        return {f[0]: getattr(self._state, f[0]) for f in self._state._fields_}

    # -----
    # _read_all
    # -----
    def _read_all(self) -> List[dict]:
        qs = self.queueSize
        data = (c.POINTER(fxe.deviceStateDicts[self._deviceName]) * qs)()
        allData = []

        nRead = self._clib.all_read(self.deviceId, data, qs)

        try:
            assert nRead == qs
        except AssertionError as err:
            raise RuntimeError("Could not read all data.") from err

        for i in range(nRead):
            allData.append({f[0]: getattr(data[i], f[0]) for f in data[i]._fields_})

        return allData

    # -----
    # rigidVersion
    # -----
    @property
    def rigidVersion(self) -> str:
        raise RuntimeError("Devices don't currently know their own hardware version.")



"""
Demo 2: Two position control

Here we demonstrate how to:
    * Command motor position
    * Read data from the device
    * Set the motor gains
"""
import platform
from time import sleep

import numpy as np
import pandas as pd
import seaborn as sns

from flexsea.device import Device


if "windows" == platform.system().lower():
    msg = "WARNING: these demos may not function properly on Windows "
    msg += "due to the way the OS handles timing. They are best run on "
    msg += "linux."
    print(msg)


# Instantiate and connect to the device; see demo1.py
device = Device()
device.open()


# In order to read data from the device, we have to instruct the device
# to send its data to the computer via the `start_streaming` method.
# This method takes in the frequency (in hertz) at which you would like
# the device to send data
frequency = 100
device.start_streaming(frequency)


# Once streaming, we use the `read` method. This method returns a
# dictionary. The fields available in the dictionary depend on the
# device being read as well as the firmware version
data = device.read()


# In this demo we are going to command the motor to move between two
# positions: the initial position and then the initial position plus
# some offset. The motor position is given in "ticks", and the
# conversion between ticks and degrees depends on the device. For an
# actpack, the conversion is 350 degrees = 16384 ticks
offset = 10000
pos0 = data["mot_ang"]
positions = [pos0, pos0 + offset]


# In order for the motor to move properly, we have to set the gains.
# The exact values of the gains depends on the device and the
# application, but here we are assuming that an actpack under no load
# is being used. The gains are set via the `set_gains` method, which
# expects the gains to be given in the following order: proportional,
# integral, differential, stiffness (impedance), damping (impedance),
# and feed-forward
gains = {
    "kp": 50,
    "ki": 0,
    "kd": 0,
    "k": 0,
    "b": 0,
    "ff": 0,
}
device.set_gains(**gains)


# Now we can command the motor to oscillate between the two positions
# several times. This is done by specifying how long we want the demo
# to last for (in seconds), how long we want to wait between position
# swaps (in seconds), and how long we want to wait between device
# reads. This allows us to finely sample the device's data for plotting
# purposes
runTime = 10
transitionTime = 2
commandDelay = 0.1
positionIndex = 0

measuredPosition = []
desiredPosition = []
deviceTime = []

nLoops = int(runTime / commandDelay)
transitionSteps = int(transitionTime / commandDelay)

# Oscillate between the two positions nLoops times
for i in range(nLoops):
    data = device.read()

    # Save the data for plotting
    measuredPosition.append(data["mot_ang"])
    desiredPosition.append(positions[positionIndex])
    deviceTime.append(data["state_time"])

    # Every transitionTime seconds, we go to the next position
    if i % transitionSteps == 0:
        positionIndex = (positionIndex + 1) % len(positions)
        # Here we tell the motor which position to move to, with the
        # value we give it in ticks
        device.command_motor_position(int(positions[positionIndex]))

    # Wait the desired amount of time for the device to process and
    # complete the command
    sleep(commandDelay)

device.close()


# Now we can plot the data to see how well the measured motor position
# tracked the desired square wave. First we put the data into a
# data frame with three columns: time, position, and position type.
# Position type indicates whether the value in the positio column is
# a measured position or desired position and makes plotting these
# values on the same plot much easier via seaborn's `hue` keyword
nValues = len(deviceTime)
t = np.concatenate((deviceTime, deviceTime))
positionType = ["Desired Position"] * nValues + ["Measured Position"] * nValues
position = np.concatenate((desiredPosition, measuredPosition))

data = {
    "Time (ms)": t,
    "Position Type": positionType,
    "Motor Position (ticks)": position,
}

df = pd.DataFrame(data)

plot = sns.relplot(
    data=df, x="Time (ms)", y="Motor Position (ticks)", hue="Position Type", kind="line"
)

plot.figure.savefig("two_position_demo.png")

print("Data plot saved as: 'two_position_demo.png'")



import ctypes as c
import sys
from time import sleep
from typing import List

import semantic_version as sem

from . import enums as fxe
from . import utilities as fxu


# ============================================
#                 DephyDevice
# ============================================
class DephyDevice:
    """
    Representation of one of Dephy's devices.
    """

    UNDEFINED = fxe.dephyDeviceErrorCodes["UNDEFINED"]
    SUCCESS = fxe.dephyDeviceErrorCodes["SUCCESS"]
    FAILURE = fxe.dephyDeviceErrorCodes["FAILURE"]
    INVALID_PARAM = fxe.dephyDeviceErrorCodes["INVALID_PARAM"]
    INVALID_DEVICE = fxe.dephyDeviceErrorCodes["INVALID_DEVICE"]
    NOT_STREAMING = fxe.dephyDeviceErrorCodes["NOT_STREAMING"]

    # -----
    # constructor
    # -----
    def __init__(
        self,
        port: str,
        baudRate: int,
        cLibVersion: str,
        logLevel: int,
        loggingEnabled: bool,
        libFile: str = "",
    ) -> None:
        self.port = port
        self.baudRate = baudRate
        self.cLibVersion = cLibVersion
        self.logLevel = logLevel
        self.loggingEnabled = loggingEnabled
        self.libFile = libFile

        self.fields: List = []
        self.deviceId: int = self.INVALID_DEVICE.value
        self.hasHabs: bool = False
        self.isChiral: bool = False
        self.streamingFrequency: int = 0
        self.heartbeatPeriod: int = 0
        self.useSafety: bool = False
        self._deviceName: str = ""
        self._deviceSide: str = ""
        self._closed: bool = False

        self._clib = fxu.load_clib(self.cLibVersion, libFile=self.libFile)

    # -----
    # destructor
    # -----
    def __del__(self) -> None:
        if not self._closed:
            self.close()

    # -----
    # isOpen
    # -----
    @property
    def isOpen(self) -> bool:
        return self._clib.is_open(self.deviceId)

    # -----
    # isStreaming
    # -----
    @property
    def isStreaming(self) -> bool:
        return self._clib.is_streaming(self.deviceId)

    # -----
    # deviceName
    # -----
    @property
    def deviceName(self) -> str:
        """
        Queries the device for it's type name, e.g., `actpack`.

        Raises
        ------
        RuntimeError:
            If we cannot get the device's name.
        """
        if self._deviceName:
            return self._deviceName

        maxDeviceNameLength = self._clib.get_max_device_name_length()
        deviceName = (c.c_char * maxDeviceNameLength)()

        if self._clib.get_device_name(self.deviceId, deviceName) != self.SUCCESS.value:
            raise RuntimeError("Could not get device name.")

        return deviceName.value.decode("utf8")

    # -----
    # deviceSide
    # -----
    @property
    def deviceSide(self) -> str:
        """
        Queries the device for it's side name, e.g., `left`.

        Raises
        ------
        RuntimeError:
            If we cannot get the device's side.
        """
        if self._deviceSide:
            return self._deviceSide

        maxDeviceSideLength = self._clib.get_max_device_side_length()
        deviceSide = (c.c_char * maxDeviceSideLength)()

        if self._clib.get_side(self.deviceId, deviceSide) != self.SUCCESS.value:
            raise RuntimeError("Could not get device name.")

        side = deviceSide.value.decode("utf8")

        # If side isn't applicable (for, e.g., an actpack), string is empty
        return side if side else "undefined"

    # -----
    # open
    # -----
    def open(self) -> None:
        """
        Establish a connection with a device.

        Raises
        ------
        IOError:
            If we fail to open the device.
        """
        self._open()
        self._setup()
        self._closed = False

    # -----
    # _open
    # -----
    def _open(self) -> None:
        if self.isOpen:
            return

        if not self.port:
            print("No port given. Searching...")
            self.port = fxu.find_port(self.baudRate, self.cLibVersion, self.libFile)
            if self.port:
                print(f"Found device at: {self.port}")
            else:
                raise IOError("Could not find a device.")

        port = self.port.encode("utf-8")
        self.deviceId = self._clib.open(port, self.baudRate, self.logLevel)

        if self.deviceId in (self.INVALID_DEVICE.value, -1):
            raise IOError("Failed to open device.")

    # -----
    # _setup
    # -----
    def _setup(self) -> None:
        givenVer = sem.Version(self.cLibVersion)
        libVer = sem.Version(self.libsVersion)

        try:
            assert givenVer == libVer
        except AssertionError as err:
            if givenVer.major == libVer.major:
                msg = f"Given lib version: `{givenVer}` doesn't match file lib "
                msg += f"version: `{libVer}`, but major versions match. Proceed[y|n]?"
                proceed = input(msg)
                if proceed != "y":
                    sys.exit(1)
            else:
                msg = f"{givenVer} doesn't match {libVer} (C lib version)"
                raise AssertionError(msg) from err

        self._deviceName = self.deviceName
        self._deviceSide = self.deviceSide

        print("Check hasHabs to make sure correct device name is present.")
        if self._deviceName in fxe.hasHabs:
            self.hasHabs = True

        if self._deviceName in fxe.hasChirality:
            self.isChiral = True

        self.fields = self._get_fields()

    # -----
    # libsVersion
    # -----
    @property
    def libsVersion(self) -> str:
        """
        Gets the version of the precompiled C libraries being used.

        Raises
        ------
        RuntimeError:
            If we fail to get the version.
        """
        major = c.c_uint16(-1)
        minor = c.c_uint16(-1)
        patch = c.c_uint16(-1)

        retCode = self._clib.get_libs_version(
            c.byref(major), c.byref(minor), c.byref(patch)
        )

        if retCode != self.SUCCESS.value:
            raise RuntimeError("Could not determine clibs version.")

        return f"{major.value}.{minor.value}.{patch.value}"

    # -----
    # _get_fields
    # -----
    def _get_fields(self) -> List[str]:
        """
        Query the device for its available data fields.

        Raises
        ------
        RuntimeError
            If we fail to get the device's field names.
        """
        if self.fields:
            return self.fields

        maxFields = self._clib.get_max_data_elements()
        maxFieldLength = self._clib.get_max_field_name_length()
        nLabels = c.c_int()

        # https://code.activestate.com/lists/python-list/704158
        labels = (c.POINTER(c.c_char) * maxFields)()

        for i in range(maxFields):
            labels[i] = c.create_string_buffer(maxFieldLength)

        retCode = self._clib.get_fields(self.deviceId, labels, c.byref(nLabels))

        if retCode != self.SUCCESS.value:
            raise RuntimeError("Could not get device field labels.")

        # Convert the labels from chars to python strings
        fields = [""] * nLabels.value
        for i in range(nLabels.value):
            for j in range(maxFieldLength):
                fields[i] += labels[i][j].decode("utf8")
            fields[i].strip("\x00")

        return fields

    # -----
    # close
    # -----
    def close(self) -> None:
        """
        Disconnect from a device.
        """
        if self.isStreaming:
            self.stop_streaming()

        if self.isOpen:
            self.stop_motor()
            self._clib.close(self.deviceId)

        # If close is called manually, when destructor is called during
        # garbage collection, isStreaming will raise a runtimeerror
        self._closed = True

    # -----
    # start_streaming
    # -----
    def start_streaming(
        self, frequency: int, heartbeatPeriod: int = 50, useSafety: bool = False
    ) -> None:
        """
        Start streaming data from a device.

        Parameters
        ----------
        frequency : int
            The desired frequency of communication.

        heartbeatPeriod : int
            When streaming, the computer periodically sends a message to
            the device to let it know that the connection between them
            is still alive. These are called heartbeat messages. This
            variable specifies the amount of time (in milliseconds)
            between successive heartbeat messages. This is related to
            how long the device will wait without receiving a heartbeat
            before shutting itself off (five times `heartbeatPeriod`).

        useSafety : bool
            If True, the device will shut itself off if it doesn't
            receive a heartbeat message from the computer within the
            allotted time (five times `heartbeatPeriod`). If False,
            the device will not shut itself off, just stop streaming,
            if a heartbeat isn't received.

        Raises
        ------
        RuntimeError:
            If the stream failed or if `open` hasn't been called.

        ValueError:
            If the heartbeatPeriod is invalid.
        """
        if self.isStreaming:
            print("Already streaming.")
            return

        if not self.isOpen:
            raise RuntimeError("Call `open` first.")

        self.streamingFrequency = frequency
        self.heartbeatPeriod = heartbeatPeriod
        self.useSafety = useSafety

        _log = 1 if self.loggingEnabled else 0

        if self.useSafety:
            hbp = self.heartbeatPeriod

            try:
                assert 50 <= hbp < self.streamingFrequency
            except AssertionError as err:
                msg = "Heartbeat period must be >= 50 and < frequency."
                raise ValueError(msg) from err

            retCode = self._clib.start_streaming_with_safety(
                self.deviceId, frequency, _log, self.heartbeatPeriod
            )

        else:
            retCode = self._clib.start_streaming(self.deviceId, frequency, _log)

        if retCode != self.SUCCESS.value:
            raise RuntimeError("Could not start stream.")

    # -----
    # stop_streaming
    # -----
    def stop_streaming(self) -> None:
        """
        Stop streaming data from a device.

        Raises
        ------
        RuntimeError:
            If the stream failed to stop.
        """
        if self._clib.stop_streaming(self.deviceId) != self.SUCCESS.value:
            raise RuntimeError("Failed to stop streaming.")

    # -----
    # read
    # -----
    def read(self, allData: bool = False) -> dict | List[dict]:
        """
        Reads data from a streaming device.

        Parameters
        ----------
        allData : bool
            If `True`, read the entire queue. If `False`, read only the
            most recent data.

        Raises
        ------
        RuntimeError:
            If not streaming.

        Returns
        -------
        dict:
            Data read from device.
        """
        if not self.isStreaming:
            raise RuntimeError("Must call `start_streaming()` before reading data.")

        return self._read() if not allData else self._read_all()

    # -----
    # _read
    # -----
    def _read(self) -> dict:
        """
        The device returns a list of values. We then have to pair those values
        with their corresponding labels.

        Raises
        ------
        RuntimeError:
            If reading fails.

        AssertionError:
            If the number of fields and amount of data read differ.

        Returns
        -------
        dict:
            Label-value pairs read from the device.
        """
        maxDataElements = self._clib.get_max_data_elements()
        nFields = c.c_int()
        deviceData = (c.POINTER(c.c_uint32) * maxDataElements)()

        retCode = self._clib.read(self.deviceId, deviceData, c.byref(nFields))

        if retCode != self.SUCCESS.value:
            raise RuntimeError("Could not read from device.")

        try:
            assert nFields.value == len(self.fields)
        except AssertionError as err:
            raise AssertionError("Incorrect number of fields read.") from err

        data = [deviceData[i].value for i in range(nFields.value)]

        return dict(zip(self.fields, data))

    # -----
    # _read_all
    # -----
    def _read_all(self) -> List[dict]:
        """
        Data from each timestep is stored in a queue on the device. Here
        we get the current size of that queue and then read from it.

        NOTE: Can the queue size change between getting the size and reading
        the data from it?

        Raises
        ------
        AssertionError:
            If the number of fields differs from what's expected.

        Returns
        -------
        List[dict]:
            Each element in the list is data from a particular timestep keyed
            by the field name.
        """
        qs = self.queueSize
        data = (c.POINTER(c.c_uint32) * qs)()
        nElements = c.c_int()

        for i in range(qs):
            data[i] = (c.c_uint32 * len(self.fields))()

        self._clib.read_all(self.deviceId, data, c.byref(nElements))

        try:
            assert nElements.value == len(self.fields)
        except AssertionError as err:
            raise AssertionError(
                "Different number of fields read than expected."
            ) from err

        allData = []

        for i in range(qs):
            singleTimeStepData = []
            for j in range(nElements.value):
                singleTimeStepData.append(data[i][j])
            allData.append(dict(zip(self.fields, singleTimeStepData)))

        return allData

    # -----
    # set_gains
    # -----
    def set_gains(self, kp: int, ki: int, kd: int, k: int, b: int, ff: int) -> None:
        """
        Sets the gains used by PID controllers on the device.

        Parameters
        ----------
        kp : int
            Proportional gain.

        ki : int
            Integral gain.

        kd : int
            Differential gain.

        k : int
            Stiffness (used in impedence control only).

        b : int
            Damping (used in impedance control only).

        ff : int
            Feed forward gain.

        Raises
        ------
        RuntimeError:
            Command failed.
        """
        devId = self.deviceId
        if self._clib.set_gains(devId, kp, ki, kd, k, b, ff) != self.SUCCESS.value:
            raise RuntimeError("Command failed")

    # -----
    # command_motor_position
    # -----
    def command_motor_position(self, value: int) -> None:
        """
        Sets motor to given position.

        Parameters
        ----------
        value : int
            Desired motor position in encoder units.

        Raises
        ------
        RuntimeError:
            If the command failed.
        """
        devId = self.deviceId
        controller = fxe.controllers["position"]
        if (
            self._clib.send_motor_command(devId, controller, value)
            != self.SUCCESS.value
        ):
            raise RuntimeError("Coult not command motor position.")

    # -----
    # command_motor_current
    # -----
    def command_motor_current(self, value: int) -> None:
        """
        Sends the given current to the motor.

        Parameters
        ----------
        value : int
            Desired motor current in milli-Amps.

        Raises
        ------
        RuntimeError:
            If the command failed.
        """
        devId = self.deviceId
        controller = fxe.controllers["current"]
        if (
            self._clib.send_motor_command(devId, controller, value)
            != self.SUCCESS.value
        ):
            raise RuntimeError("Coult not command motor current.")

    # -----
    # command_motor_voltage
    # -----
    def command_motor_voltage(self, value: int) -> None:
        """
        Sets motor's voltage.

        Parameters
        ----------
        value : int
            Desired motor voltage in milli-volts.

        Raises
        ------
        RuntimeError:
            If the command failed.
        """
        devId = self.deviceId
        controller = fxe.controllers["voltage"]
        if (
            self._clib.send_motor_command(devId, controller, value)
            != self.SUCCESS.value
        ):
            raise RuntimeError("Coult not command motor voltage.")

    # -----
    # command_motor_impedance
    # -----
    def command_motor_impedance(self, value: int) -> None:
        """
        Has the motor simulate a stretched from current position to
        the desired position using the damping gain b and stiffness
        gain k.

        Parameters
        ----------
        value : int
            Desired motor position in ticks.

        Raises
        ------
        RuntimeError:
            If the command failed.
        """
        devId = self.deviceId
        controller = fxe.controllers["impedance"]
        if (
            self._clib.send_motor_command(devId, controller, value)
            != self.SUCCESS.value
        ):
            raise RuntimeError("Coult not command motor impedance.")

    # -----
    # stop_motor
    # -----
    def stop_motor(self):
        """
        Stops the motor.

        Raises
        ------
        RuntimeError:
            If the command failed.
        """
        devId = self.deviceId
        controller = fxe.controllers["none"]
        if self._clib.send_motor_command(devId, controller, 0) != self.SUCCESS.value:
            raise RuntimeError("Could not stop motor.")

    # -----
    # find_poles
    # -----
    def find_poles(self) -> None:
        """
        DO NOT USE THIS FUNCTION UNLESS YOU KNOW WHAT YOU ARE DOING!

        Instructs the device to go through the pole-finding process.

        Raises
        ------
        ValueError:
            If the command failed.
        """
        userInput = input(
            "WARNING: You should not use this function unless you know what "
            "you are doing!\nProceed?[y/n] "
        )

        if userInput != "y":
            print("Aborting pole finding.")
            return

        if self._clib.find_poles(self.deviceId) != self.SUCCESS.value:
            raise ValueError("Command failed")

        msg = "NOTE: Please wait for the process to complete. The motor will stop "
        msg += "moving when it is done. It usually takes 1-2 minutes to complete."
        print(msg)

        msg = "NOTE: Once the pole-finding procedure is completed, you must "
        msg += "power cylce the device for the changes to take effect."
        print(msg)

    # -----
    # activate_bootloader
    # -----
    def activate_bootloader(self, target: str) -> None:
        """
        Activates target bootloader.

        Parameters
        ----------
        target : str
            Bootloader target. Can be: mn, ex, re, or habs.

        Raises
        ------
        RuntimeError:
            Command failed.
        """
        targetCode = fxe.bootloaderTargets[target]

        returnCode = self._clib.activate_bootloader(self.deviceId, targetCode)

        if returnCode != self.SUCCESS.value:
            raise RuntimeError(f"Could not activate bootloader for: `{target}`.")

    # -----
    # bootloaderActive
    # -----
    @property
    def bootloaderActive(self) -> bool:
        """
        Get status of bootloader.

        Raises
        ------
        RunttimeError:
            Command failed.

        Returns
        -------
        bool
            `True` if bootloader is active and `False` otherwise.
        """
        if self._clib.is_bootloader_activated(self.deviceId) != self.SUCCESS.value:
            return False

        return True

    # -----
    # rigidVersion
    # -----
    @property
    def rigidVersion(self) -> str:
        raise RuntimeError("Devices don't currently know their own hardware version.")

    # -----
    # firmware
    # -----
    @property
    def firmware(self) -> dict:
        """
        Gets the fimware versions of device's MCUs.

        Raises
        ------
        RuntimeError:
            Command failed.

        Returns
        -------
        Dict
            A dictionary with the semantic version strings of manage,
            execute, and regulate's firmware. And habs, if applicable.
        """
        if self._clib.request_firmware_version(self.deviceId) != self.SUCCESS.value:
            raise RuntimeError("Command failed")

        sleep(5)

        fw = self._clib.get_last_received_firmware_version(self.deviceId)
        fwDict = {
            "mn": fxu.decode(fw.mn),
            "ex": fxu.decode(fw.ex),
            "re": fxu.decode(fw.re),
        }

        if self.hasHabs:
            fwDict["habs"] = fxu.decode(fw.habs)

        return fwDict

    # -----
    # print
    # -----
    def print(self, data: dict | None = None) -> None:
        """
        Reads the data from the device and then prints it to the screen.
        If data is given, we print that instead.
        """
        if data is None:
            data = self.read()
        for key, value in data.items():
            print(f"{key} : {value}")

    # -----
    # queueSize
    # -----
    @property
    def queueSize(self) -> int:
        """
        Get the maximum read data queue size of a device.

        Returns
        -------
        int:
            Maximum read data queue size of a device.
        """
        return self._clib.get_read_data_queue_size(self.deviceId)

    @queueSize.setter
    def queueSize(self, dataSize: int) -> None:
        """
        Sets the maximum read data queue size of a device.

        Parameters
        ----------
        dataSize : int
            Size to set the read data queue size to.

        Raises
        ------
        ValueError:
            If data size is invalid.

        RuntimeError:
            If the command failed.
        """
        returnCode = self._clib.set_read_data_queue_size(self.deviceId, dataSize)

        if returnCode == self.INVALID_PARAM:
            raise ValueError(f"Invalid data_size: {dataSize}")
        if returnCode == self.FAILURE:
            raise RuntimeError("Command failed")

    # -----
    # set_tunnel_mode
    # -----
    def set_tunnel_mode(self, target: str, timeout: int = 30) -> bool:
        """
        All communication goes through Manage, so we need to put it into
        tunnel mode in order to activate the other bootloaders. When bootloading
        Manage itself, this causes it to reboot in DFU mode.

        Parameters
        ----------
        target : str
            The name of the target to set (abbreviated).

        timeout : int
            The number of seconds to wait for confirmation before failing.

        Raises
        ------
        IOError
            If the device cannot be opened.

        Returns
        -------
        result : bool
            If `True`, the bootloader was set successfully. If `False` then
            something went wrong.
        """
        if not self.isOpen:
            raise IOError("Error: device must be open before setting tunnel mode.")

        activated = False

        while timeout > 0 and not activated:
            if timeout % 5 == 0:
                try:
                    self.activate_bootloader(target)
                except IOError:
                    pass

            # Device gets disconnected briefly when Mn resets, so we wait
            sleep(1)
            timeout -= 1
            activated = self.bootloaderActive

        return activated

    # -----
    # uvlo
    # -----
    @property
    def uvlo(self) -> int:
        """
        Gets the currently set UVLO.
        """
        if self._clib.request_uvlo(self.deviceId) != self.SUCCESS.value:
            raise RuntimeError("Could not request firmware version.")

        # Let the device process the request
        sleep(5)

        uvlo = self._clib.read_uvlo(self.deviceId)

        if uvlo == -1:
            raise RuntimeError("Could not get requested UVLO.")

        return uvlo

    @uvlo.setter
    def uvlo(self, value: int) -> None:
        """
        Sets the UVLO value for the device. `value` needs to be in milli-volts.
        """
        if self._clib.set_uvlo(self.deviceId, value) != self.SUCCESS.value:
            raise RuntimeError("Could not set UVLO.")

    # -----
    # calibrate_imu
    # -----
    def calibrate_imu(self) -> None:
        """
        DO NOT USE THIS FUNCTION UNLESS YOU KNOW WHAT YOU ARE DOING!

        Instructs the device to go through the IMU calibration process.

        Raises
        ------
        ValueError:
            If the command failed.
        """
        userInput = input(
            "WARNING: You should not use this function unless you know what "
            "you are doing!\nProceed?[y/n] "
        )

        if userInput != "y":
            print("Aborting IMU calibration.")
            return

        if self._clib.calibrate_imu(self.deviceId) != self.SUCCESS.value:
            raise RuntimeError("Could not calibrate imu.")



import ctypes as c

from flexsea import firmware as fw
from flexsea.specs.device_specs import all_devices as fxd


v10_0_0 = {
    "commands": {
        "is_open": {
            "name": "fxIsOpen",
            "argTypes": [
                c.c_uint,
            ],
            "returnType": c.c_bool,
        },
        "is_streaming": {
            "name": "fxIsStreaming",
            "argTypes": [
                c.c_uint,
            ],
            "returnType": c.c_bool,
        },
        "get_max_device_name_length": {
            "name": "fxGetMaxDeviceNameLength",
            "argTypes": [],
            "returnType": c.c_int,
        },
        "get_device_name": {
            "name": "fxGetDeviceTypeNameWrapper",
            "argTypes": [c.c_uint, c.c_char_p],
            "returnType": c.c_int,
        },
        "get_max_device_side_length": {
            "name": "fxGetMaxDeviceSideNameLength",
            "argTypes": [],
            "returnType": c.c_int,
        },
        "get_side": {
            "name": "fxGetDeviceSideNameWrapper",
            "argTypes": [c.c_uint, c.c_char_p],
            "returnType": c.c_int,
        },
        "get_libs_version": {
            "name": "fxGetLibsVersion",
            "argTypes": [
                c.POINTER(c.c_uint16),
                c.POINTER(c.c_uint16),
                c.POINTER(c.c_uint16),
            ],
            "returnType": c.c_int,
        },
        "get_max_field_name_length": {
            "name": "fxGetMaxDataLabelLength",
            "argTypes": [],
            "returnType": c.c_int,
        },
        "get_fields": {
            "name": "fxGetDataLabelsWrapper",
            "argTypes": [c.c_uint, c.POINTER(c.POINTER(c.c_char)), c.POINTER(c.c_int)],
            "returnType": c.c_int,
        },
        "get_max_data_elements": {
            "name": "fxGetMaxDataElements",
            "argTypes": [],
            "returnType": c.c_int,
        },
        "read": {
            "name": "fxReadDevice",
            "argTypes": [c.c_uint, c.POINTER(c.c_uint32), c.POINTER(c.c_int)],
            "returnType": c.c_int,
        },
        "read_all": {
            "name": "fxReadDeviceAllWrapper",
            "argTypes": [c.c_uint, c.POINTER(c.POINTER(c.c_int)), c.POINTER(c.c_int)],
            "returnType": None,
        },
        "request_uvlo": {
            "name": "fxRequestUVLO",
            "argTypes": [
                c.c_uint,
            ],
            "returnType": c.c_int,
        },
        "read_uvlo": {
            "name": "fxGetLastReceivedUVLO",
            "argTypes": [
                c.c_uint,
            ],
            "returnType": c.c_int,
        },
        "set_uvlo": {
            "name": "fxSetUVLO",
            "argTypes": [c.c_uint, c.c_uint],
            "returnType": c.c_int,
        },
        "open": {
            "name": "fxOpen",
            "argTypes": [c.c_char_p, c.c_uint, c.c_uint],
            "returnType": c.c_int,
        },
        "close": {
            "name": "fxClose",
            "argTypes": [
                c.c_uint,
            ],
            "returnType": c.c_int,
        },
        "start_streaming": {
            "name": "fxStartStreaming",
            "argTypes": [c.c_uint, c.c_uint, c.c_bool],
            "returnType": c.c_int,
        },
        "start_streaming_with_safety": {
            "name": "fxStartStreamingWithSafety",
            "argTypes": [c.c_uint, c.c_uint, c.c_bool, c.c_uint16],
            "returnType": c.c_int,
        },
        "stop_streaming": {
            "name": "fxStopStreaming",
            "argTypes": [
                c.c_uint,
            ],
            "returnType": c.c_int,
        },
        "set_gains": {
            "name": "fxSetGains",
            "argTypes": [
                c.c_uint,
                c.c_uint,
                c.c_uint,
                c.c_uint,
                c.c_uint,
                c.c_uint,
                c.c_uint,
            ],
            "returnType": c.c_int,
        },
        "send_motor_command": {
            "name": "fxSendMotorCommand",
            "argTypes": [c.c_uint, c.c_int, c.c_int],
            "returnType": c.c_int,
        },
        "find_poles": {
            "name": "fxFindPoles",
            "argTypes": [
                c.c_uint,
            ],
            "returnType": c.c_int,
        },
        "activate_bootloader": {
            "name": "fxActivateBootloader",
            "argTypes": [c.c_uint, c.c_uint8],
            "returnType": c.c_int,
        },
        "is_bootloader_activated": {
            "name": "fxIsBootloaderActivated",
            "argTypes": [
                c.c_uint,
            ],
            "returnType": c.c_int,
        },
        "request_firmware_version": {
            "name": "fxRequestFirmwareVersion",
            "argTypes": [
                c.c_uint,
            ],
            "returnType": c.c_int,
        },
        "get_last_received_firmware_version": {
            "name": "fxGetLastReceivedFirmwareVersion",
            "argTypes": [
                c.c_uint,
            ],
            "returnType": fw.Firmware,
        },
        "get_read_data_queue_size": {
            "name": "fxGetReadDataQueueSize",
            "argTypes": [
                c.c_uint,
            ],
            "returnType": c.c_int,
        },
        "set_read_data_queue_size": {
            "name": "fxSetReadDataQueueSize",
            "argTypes": [c.c_uint, c.c_uint],
            "returnType": c.c_uint,
        },
    },
}



# pylint: skip-file
"""
///
/// @file BattCyclerState.py
///
/// @brief AUTOGENERATED FILE! ONLY EDIT IF YOU ARE A MACHINE!
///
/// @core 0bad34709094958919bfe30e6c0ca13c7682c49a
///
/// @build df4ac791e9dd25df235f8ed979101d1d2e75b19b
///
/// @date 2020-09-28 18:37:52
///
/// @author Dephy, Inc.
"""
from ctypes import Structure, c_int


class BattCyclerState(Structure):
    _pack_ = 1
    _fields_ = [
        ("battcycler", c_int),
        ("id", c_int),
        ("state_time", c_int),
        ("current", c_int),
        ("voltage", c_int),
        ("esr", c_int),
        ("fsm_state", c_int),
        ("button", c_int),
        ("leds", c_int),
        ("genvar_0", c_int),
        ("genvar_1", c_int),
        ("genvar_2", c_int),
        ("genvar_3", c_int),
        ("status", c_int),
        ("test_in_progress_now", c_int),
        ("last_completed_test", c_int),
        ("result_of_last_test", c_int),
        ("time_elapsed_in_test", c_int),
        ("error_codes", c_int),
        ("charge_discharge_cycles_completed", c_int),
        ("charge_discharge_total_cycles", c_int),
        ("desired_battery_current", c_int),
        ("SystemTime", c_int),
    ]



# pylint: skip-file
"""
///
/// @file NetNodeState.py
///
/// @brief AUTOGENERATED FILE! ONLY EDIT IF YOU ARE A MACHINE!
///
/// @core 0bad34709094958919bfe30e6c0ca13c7682c49a
///
/// @build df4ac791e9dd25df235f8ed979101d1d2e75b19b
///
/// @date 2020-06-22 14:42:45
///
/// @author Dephy, Inc.
"""
from ctypes import Structure, c_int


class NetNodeState(Structure):
    _pack_ = 1
    _fields_ = [
        ("netnode", c_int),
        ("id", c_int),
        ("state_time", c_int),
        ("genvar_0", c_int),
        ("genvar_1", c_int),
        ("genvar_2", c_int),
        ("genvar_3", c_int),
        ("status", c_int),
        ("SystemTime", c_int),
    ]



import ctypes as c
import hashlib
import os
from pathlib import Path
import platform
import sys

import boto3
from botocore.client import BaseClient
import botocore.exceptions as bce
from serial.tools.list_ports import comports

from flexsea.specs.api_spec import apiSpec
import flexsea.enums as fxe

from . import config as cfg


# ============================================
#                   get_os
# ============================================
def get_os() -> str:
    """
    Returns the operating system and "bitness":

        * windows_32bit
        * windows_64bit
        * pi_32bit
        * pi_64bit
        * linux_32bit
        * linux_64bit

    Returns
    -------
    os : str
        The name and "bitness" of the current operating system.
    """
    system = platform.system().lower()

    if system == "linux":
        machine = os.uname().machine
        if machine.startswith("arm") or machine.startswith("aarch"):
            system = "pi"

    return system + "_" + platform.architecture()[0]


# ============================================
#                    decode
# ============================================
def decode(val: int) -> str:
    """
    Returns decoded version number formatted as x.y.z
    """
    major: int = 0
    minor: int = 0
    bug: int = 0

    if val > 0:
        while val % 2 == 0:
            major += 1
            val = int(val / 2)

        while val % 3 == 0:
            minor += 1
            val = int(val / 3)

        while val % 5 == 0:
            bug += 1
            val = int(val / 5)

    return f"{major}.{minor}.{bug}"


# ============================================
#                  load_clib
# ============================================
def load_clib(
    cLibVersion: str, silent: bool = False, libFile: str | Path = ""
) -> c.CDLL:
    """
    Uses `ctypes` to load the appropriate C libraries depending on the
    OS.

    Parameters
    ----------
    cLibVersion : str
        The version of the pre-compiled libraries to use. The major
        version should match the major version of the firmware on the
        device. If no libraries are found, then we download them from
        AWS.

    silent : bool (optional)
        If `True`, suppresses messages to stdout.

    libFile : str (optional)
        If set, specifies the library file to load. Overrides looking
        in the standard .dephy location and will not attempt to
        download the file if it isn't found. Still requires cLibVersion
        to know what api to use with the lib.

    Raises
    ------
    OSError:
        Raised if the precompiled libraries fail to load.

    Returns
    -------
    clib : ctypes.cdll.CDLL
        The Python object from which we can call the flexsea C
        functions.
    """
    _os = get_os()

    if not libFile:
        lib = cfg.windowsLib if "win" in _os else cfg.linuxLib
        libFile = cfg.libsDir.joinpath(cLibVersion, _os, lib)

        if not libFile.exists():
            libFile.parent.mkdir(parents=True, exist_ok=True)
            libObj = str(Path("libs").joinpath(cLibVersion, _os, lib).as_posix())

            download(libObj, cfg.libsBucket, str(libFile))
    else:
        libFile = Path(libFile)

    if "win" in _os:
        try:
            for extraPath in os.environ["PATH"].split(";"):
                if os.path.exists(extraPath) and "mingw" in extraPath:
                    os.add_dll_directory(extraPath)
            os.add_dll_directory(str(libFile.parent.absolute()))
        except OSError:
            msg = f"Error loading precompiled library: `{libFile}`\n"
            msg += "The most likely cause is a mismatch between the Python, pip and "
            msg += "shell architectures.\nPlease ensure all three are either 32 or 64 "
            msg += "bit.\nKeep different versions isolated by virtual environments.\n"
            print(msg)
            sys.exit(1)

    clib = c.cdll.LoadLibrary(str(libFile))

    api = apiSpec[cLibVersion]

    for functionName, functionData in api["commands"].items():
        func = getattr(clib, functionData["name"], None)
        if func:
            func.argtypes = functionData["argTypes"]
            func.restype = functionData["returnType"]
        setattr(clib, functionName, func)

    if not silent:
        print(f"Using version: {cLibVersion} of pre-compiled C library.")
        print(f"Loading library from: {libFile}")

    return clib


# ============================================
#                  download
# ============================================
def download(fileObj: str, bucket: str, dest: str, profile: str | None = None) -> None:
    """
    Downloads `fileObj` from `bucket` to `dest` with the AWS
    credentials profile `profile`.

    Raises
    ------
    botocore.exceptions.ProfileNotFound
        If the given profile does not exist in the AWS credentials file.
        If the ~/.aws/credentials file does not exist.

    botocore.exceptions.PartialCredentialsError
        If the given profile is missing one or more required keys.

    AssertionError
        If the download fails.
    """
    try:
        session = boto3.Session(profile_name=profile)
    except bce.ProfileNotFound as err:
        raise err

    try:
        client = session.client("s3")
    except bce.PartialCredentialsError as err:
        raise err

    client.download_file(bucket, fileObj, dest)

    _validate_download(client, bucket, fileObj, dest)


# ============================================
#             _validate_download
# ============================================
def _validate_download(
    client: BaseClient, bucket: str, fileObj: str, dest: str
) -> None:
    """
    Compares the AWS md5 hash to the local md5 hash to make sure the
    files are the same.

    S3 objects have an attribute called 'ETag', which is a string.
    There are two possibilities: the string is a hex digest or the
    string is a hex digest + -NUMBER.

    In the case that ETag has no -NUMBER suffix, it means that the ETag
    is the md5 hash of the file contents.

    In the case that there is a -NUMBER suffix, it means that the file
    was uploaded to S3 in NUMBER chunks and that the hex digest is actually
    the hex digest of the digests of all of the chunks concatenated together.

    Parameters
    ----------
    client : BaseClient
        The object that allows use to communicate with S3.

    bucket : str
        The name of the bucket the file came from.

    fileObj : str
        The name of the fileObj we are validating.

    dest : str
        The name of the dowloaded file on disk.
    """
    assert Path(dest).exists()

    # Check the local file's integrity by comparing its md5 hash to
    # AWS's md5 hash, called ETag
    objData = client.head_object(Bucket=bucket, Key=fileObj)
    etag = objData["ETag"].strip('"')

    try:
        nChunks = int(etag.split("-")[1])
    except IndexError:
        nChunks = 1

    if nChunks == 1:
        with open(dest, "rb") as fd:
            data = fd.read()
        localHash = hashlib.md5(data).hexdigest()

    elif nChunks > 1:
        chunkHashes = []
        with open(dest, "rb") as fd:
            for chunk in range(1, nChunks + 1):
                objData = client.head_object(
                    Bucket=bucket, Key=fileObj, PartNumber=chunk
                )
                chunkSize = objData["ContentLength"]
                data = fd.read(chunkSize)
                if data:
                    chunkHashes.append(hashlib.md5(data))
                else:
                    break

        if len(chunkHashes) == 1:
            localHash = chunkHashes[0].hexdigest()
        else:
            digests = b"".join([m.digest() for m in chunkHashes])
            digestsMd5 = hashlib.md5(digests)
            localHash = f"{digestsMd5.hexdigest()}-{len(chunkHashes)}"

    assert localHash == etag


# ============================================
#                   find_port
# ============================================
def find_port(baudRate: int, cLibVersion: str, libFile: str = "") -> str:
    """
    Tries to establish a connection to the Dephy device given by
    the user-supplied port. If no port is supplied, then we loop
    over all available serial ports to try and find a valid device.

    Parameters
    ----------
    baudRate : int
        Baud rate for communicating with the device.

    cLibVersion : str
        The semantic version string of the firmware currently on the device.

    libFile : str (optional)
        If set, specifies the library file to load. Overrides looking
        in the standard .dephy location and will not attempt to
        download the file if it isn't found. Still requires cLibVersion
        to know what api to use with the lib.


    Raises
    ------
    RuntimeError
        If no valid Dephy device can be found.

    Returns
    -------
    devicePort : str
        Name of the device's COM port.
    """
    devicePort = None
    clib = load_clib(cLibVersion, True, libFile)

    for _port in comports():
        p = _port.device
        deviceID = clib.open(p.encode("utf-8"), baudRate, 0)
        if deviceID in (
            fxe.dephyDeviceErrorCodes["INVALID_DEVICE"].value,
            fxe.legacyDeviceErrorCodes["INVALID_DEVICE"].value,
            -1,
        ):
            continue
        devicePort = p
        clib.close(deviceID)
        break

    if not devicePort:
        raise RuntimeError("Could not find a valid device.")

    return devicePort



import ctypes as c


# ============================================
#                  Firmware
# ============================================
class Firmware(c.Structure):
    """
    Holds the integer values representing the firmware versions of
    each microcontroller as returned from the C library. These need to
    be decoded before they make sense.
    """

    _fields_ = [
        ("mn", c.c_uint32),
        ("ex", c.c_uint32),
        ("re", c.c_uint32),
        ("habs", c.c_uint32),
    ]



from .ActPackPlusState import ActPackPlusState
from .ActPackState import ActPackState
from .BattCyclerState import BattCyclerState
from .BMSState import BMSState
from .CellScreenerState import CellScreenerState
from .CycleTesterState import CycleTesterState
from .EB5xState import EB5xState
from .EB603State import EB603State
from .EB604State import EB604State
from .EB60State import EB60State
from .ExoState import ExoState
from .HabsoluteState import HabsoluteState
from .MD10State import MD10State
from .MD11State import MD11State
from .MD12State import MD12State
from .MD20State import MD20State
from .NetMasterState import NetMasterState
from .NetNodeState import NetNodeState
from .NewBooState import NewBooState
from .OPORPState import OPORPState
from .RobotTrouserState import RobotTrouserState
from .XC00sState import XC00sState
from .XC00tState import XC00tState



import semantic_version as sem

from . import config as cfg
from .dephy_device import DephyDevice
from .legacy_device import LegacyDevice
from .utilities import find_port


# ============================================
#                   Device
# ============================================
class Device(LegacyDevice):
    """
    The leaf of the device inheritance tree. The purpose of this class
    is to serve as the user-facing object.
    """

    # -----
    # __new__
    # -----
    def __new__(
        cls,
        port: str = "",
        baudRate: int = cfg.baudRate,
        cLibVersion: str = cfg.LTS,
        logLevel: int = 4,
        loggingEnabled: bool = True,
        libFile: str = "",
    ) -> DephyDevice | LegacyDevice:
        inUse = sem.Version(cLibVersion)
        cutoff = sem.Version(cfg.legacyCutoff)

        if not port:
            port = find_port(baudRate, cLibVersion, libFile)

        if inUse < cutoff:
            return LegacyDevice(
                port, baudRate, cLibVersion, logLevel, loggingEnabled, libFile
            )
        return DephyDevice(
            port, baudRate, cLibVersion, logLevel, loggingEnabled, libFile
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['demos',
 'flexsea',
 'flexsea.specs',
 'flexsea.specs._api_specs',
 'flexsea.specs.device_specs']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1.26.8,<2.0.0',
 'pyserial>=3.5,<4.0',
 'semantic-version>=2.10.0,<3.0.0']

entry_points = \
{'console_scripts': ['flexsea-demos = demos.console.main:main']}

setup_kwargs = {
    'name': 'flexsea',
    'version': '10.0.0',
    'description': '',
    'long_description': '# FlexSEA\n\n\n`flexsea` is a Python package for interacting with Dephy\'s wearable robotic devices.\nIt can be used for gathering data from a device or for writing your own controller.\n\n\n## Installation\n\nIt is **strongly** recommended that you install `flexsea` in a [virtual environment](https://docs.python.org/3/library/venv.html).\nAdditionally, `flexsea` requires [Python >= 3.11](https://www.python.org/downloads/release/python-3111/), and has been tested\non Windows and Ubuntu.\n\n**NOTE**: These instructions use the `python3` executable. If you are on Windows, you\nwill need to replace `python3` -> `python`.\n\n### Pip\n\n```bash\npython3 -m pip install flexsea\n```\n\n\n### From Source\n\nIn order to install from source, you will need [git](https://git-scm.com/downloads).\n\n```bash\ngit clone https://github.com/DephyInc/Actuator-Package.git\ncd Actuator-Package/\ngit checkout v10.0.0 # Or the branch you want\npython3 -m pip install .\n```\n\n\n## Usage\n\n### Demos\n\nA good reference for what `flexsea` is capable of and how various tasks, such as\ncontrolling the device\'s motor, can be accomplished, is the collection of demo scripts\nthat live in the `demos/` directory of the repository. There are currently six demos,\nand they should be viewed in order, as each successive demo builds off of the information\npresented in the previous one.\n\n\n### API Overview\n\n#### Importing and Instantiating\nThe central object in `flexsea` is the `Device` class. For most use cases, this is the\nonly aspect of `flexsea` that you will need to interact with directly. You can import\nit into your code like so:\n\n```python\nfrom flexsea.device import Device\n```\n\nThe constructor takes five keyword arguments:\n\n```python\nclass Device(\n    port: str="",\n    baudRate: int=cfg.baudRate,\n    cLibVersion: str=cfg.LTS,\n    logLevel: int=4,\n    loggingEnabled: bool=True\n)\n```\n\n* `port`: The name of the serial port that the device is connected to. On Windows, this is typically something akin to "COM3" and on Linux it is usually something like "/dev/ttyACM0". If you do not provide a value, `flexsea` will scan through all of the available serial ports, stopping at the first valid device that it finds. This means that this keyword is typically only useful if you have more than one device connected at once.\n* `baudRate`: The baud rate used for communicating with the device. Most of Dephy\'s devices all use the same baud rate, which is set as the default value for you.\n* `cLibVersion`: `flexsea` is a wrapper around a pre-compiled C library that actually handles all of the heavy lifting of communicating with the device. This parameter allows you to specify the semantic version string of the version of this library that you would like to use. These libraries are stored in a public AWS S3 bucket. If you do not already have the version you specify installed, then `flexsea` will attempt to download it from this bucket for you. By default, the latest LTS version is selected for you. In most cases, changing this value is only necessary for bootloading.\n* `logLevel`: Under the hood, the pre-compiled C library makes use of the [spdlog](https://github.com/gabime/spdlog) logging library. This parameter controls the verbosity of the logs, with `0` being the most verbose and `6` disabling logging all together.\n* `loggingEnabled`: If set to `True` then both data and debug logs will be generated (unless `logLevel=6`). If `False`, then no logs are generated, regardless of the value of `logLevel`.\n\nTypically, all you\'ll need to do to create an instance of the object is:\n\n```python\ndevice = Device()\n```\n\n#### Connecting and Streaming\n\nOnce instantiated, you need to establish a connection between the computer and the device. This is done via the `open` method:\n\n```python\ndevice.open()\n```\n\nAdditionally, if you would like the device to send its data to the computer -- an action called *streaming* -- then you must invoke the `start_streaming` method:\n\n```python\ndevice.start_streaming(frequency)\n```\n\nwhere `frequency` is the rate (in Hertz) at which the device will send data.\n\n**NOTE**: Currently, the maximum supported frequency is 1000Hz.\n\n\n#### Reading and Printing\n\nIf you are streaming, you can get the most recent device data from the `read` method:\n\n```python\ndata = device.read()\n```\n\nWhere `data` is a dictionary. The available fields depend on the type of device as well as the firmware version. If you have not read from the device in a while, you can get all of the data that\'s currently in the device\'s internal queue by using the `allData` keyword:\n\n```python\nallData = device.read(allData=True)\n```\n\nIn this case, the return value `allData` will be a list of dictionaries, one for each time stamp.\n\nTo conveniently display the most recent data:\n\n```python\ndevice.print()\n```\n\n`print` takes an optional keyword argument called `data`, which should be a dictionary returned by `read`. This lets you display data that was read at some arbitrary point in the past.\n\n\n#### Controlling the Motor\n```python\ndevice.command_motor_current(current) # milliamps\ndevice.command_motor_position(position) # motor ticks\ndevice.command_motor_voltage(voltage) # millivolts\ndevice.command_motor_impedance(position) # motor ticks\ndevice.stop_motor()\ndevice.set_gains(kp, ki, kd, k, b, ff) # See below\n```\n\nWhen setting the gains:\n\n* `kp`: The proportional gain\n* `ki`: The integral gain\n* `kd`: The differential gain\n* `k`: The stiffness gain for impedance control\n* `b`: The damping gain for impedance control\n* `ff`: The feed-forward gain\n\n\n#### Device State\n\nYou can also introspect certain aspects of the device\'s state, depending on the firmware version you\'re running:\n\n* `isOpen` : Indicates whether or not the computer and the device are connected\n* `isStreaming`: Indicates whether or not the device is sending data\n* `deviceName`: The name of the type of the device, e.g., "actpack"\n* `deviceSide`: Either "left" or "right", if applicable; `None` otherwise. **Requires firmware >= v10.0.0**.\n* `libsVersion`: The semantic version string of the pre-compiled C library being used. **Requires >= v10.0.0 of the pre-compiled C library**.\n* `firmware`: The semantic version string of the firmware version\n* `uvlo`: Used to both get and set the device\'s UVLO in millivolts\n\n\n#### Cleaning Up\n\nWhen finished commanding the device, it is good practice to call the `close` method:\n\n```python\ndevice.close()\n```\n\nAdditionally, when done streaming, you can call the `stop_streaming` method:\n\n```python\ndevice.stop_streaming()\n```\n\n**NOTE**: `stop_streaming` is called automatically by `close`, and `close` is called automatically by the `Device` class\' destructor\n',
    'author': 'Jared',
    'author_email': 'jcoughlin@dephy.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.11.0rc1,<4.0.0',
}


setup(**setup_kwargs)



"""
Demo 6: High Stress
"""
import platform
from time import sleep
from typing import List

import pandas as pd
import numpy as np
import seaborn as sns

from flexsea.device import Device


# ============================================
#                 send_command
# ============================================
def send_command(
    cmd: dict, cmdType: str, gains: dict, setGains: bool, plotData: dict, device: Device
) -> dict:
    deviceData = device.read()

    if setGains:
        device.set_gains(**gains)

    setPoint = cmd["cur"] if cmdType == "current" else cmd["pos"]

    if cmdType == "current":
        device.command_motor_current(int(setPoint))
    else:
        device.command_motor_position(int(setPoint))

    plotData["stateTime"].append(deviceData["state_time"])
    plotData["desiredPosition"].append(cmd["pos"])
    plotData["measuredPosition"].append(deviceData["mot_ang"])
    plotData["desiredCurrent"].append(cmd["cur"])
    plotData["measuredCurrent"].append(deviceData["mot_cur"])

    return plotData


# ============================================
#                 get_device
# ============================================
def get_device(frequency: int) -> Device:
    device = Device()
    device.open()
    device.start_streaming(frequency)

    return device


# ============================================
#                  get_gains
# ============================================
def get_gains() -> dict:
    positionGains = {
        "kp": 100,
        "ki": 10,
        "kd": 0,
        "k": 0,
        "b": 0,
        "ff": 0,
    }

    currentGains = {
        "kp": 40,
        "ki": 400,
        "kd": 0,
        "k": 0,
        "b": 0,
        "ff": 128,
    }

    gains = {
        "current": currentGains,
        "position": positionGains,
    }

    return gains


# ============================================
#                 get_samples
# ============================================
def get_samples(frequency: int) -> dict:
    # Position samples
    positionAmplitude = 10000
    positionFreq = 1

    nSamples = int(frequency / positionFreq)
    x = np.linspace(-np.pi, np.pi, nSamples)
    positionSamples = positionAmplitude * np.sin(x)

    # Current samples
    currentAmplitude = 1500
    currentFreq = 5

    nSamples = int(frequency / currentFreq)
    x = np.linspace(-np.pi, np.pi, nSamples)
    currentSamples = currentAmplitude * np.sin(x)

    # Current samples: line
    value = 0
    length = 0.5

    nSamples = np.int32(length * frequency)
    currentSamplesLine = np.array([value for _ in range(nSamples)])

    # Samples
    samples = {
        "position": positionSamples,
        "currentSine": currentSamples,
        "currentLine": currentSamplesLine,
    }

    return samples


# ============================================
#                    plot
# ============================================
def plot(time: List, desired: List, measured: List, label: str) -> None:
    nValues = len(time)
    t = np.concatenate((time, time))
    dataType = ["Desired"] * nValues + ["Measured"] * nValues
    data = np.concatenate((desired, measured))

    df = pd.DataFrame({"Time (ms)": t, "Type": dataType, label: data})

    plt = sns.relplot(data=df, x="Time (ms)", y=label, hue="Type", kind="line")

    plt.figure.savefig(f"{label}_demo6.png")


# ============================================
#                    main
# ============================================
def main():
    """
    This demo alternates between a position control loop and a current
    control loop several times in a demanding fashion. The setpoints
    for each controller are generated from two different sine waves.
    """
    if "windows" == platform.system().lower():
        msg = "WARNING: these demos may not function properly on Windows "
        msg += "due to the way the OS handles timing. They are best run on "
        msg += "linux."
        print(msg)

    frequency = 1000
    delay = 1.0 / frequency
    nLoops = 3
    # Factor to determine current on "way back" to starting point
    currentAsymmetricG = 1.15

    # Instantiate the device and start streaming
    device = get_device(frequency)

    # Initialize the gains to be used for each controller
    gains = get_gains()

    # Get the samples for each controller
    samples = get_samples(frequency)

    # Initial device position
    pos0 = device.read()["mot_ang"]

    # Plotting arrays
    data = {
        "stateTime": [],
        "desiredPosition": [],
        "measuredPosition": [],
        "desiredCurrent": [],
        "measuredCurrent": [],
    }

    for i in range(nLoops):
        # Step 0
        sleep(delay)
        if i:
            command = {"cur": 0, "pos": device.read()["mot_ang"]}
        else:
            command = {"cur": 0, "pos": pos0}
        data = send_command(command, "position", gains["position"], True, data, device)

        # Step 1
        if i:
            for sample in np.linspace(data["mot_ang"], pos0, 360):
                command = {"cur": 0, "pos": sample}
                sleep(delay)
                data = send_command(command, "position", {}, False, data, device)

        # Step 2
        for sample in samples["position"]:
            command = {"cur": 0, "pos": sample + pos0}
            sleep(delay)
            data = send_command(command, "position", {}, False, data, device)

        # Step 3
        command = {"cur": 0, "pos": pos0}
        data = send_command(command, "current", gains["current"], True, data, device)
        sleep(delay)

        # Step 4
        for sample in samples["currentSine"].astype(np.int64):
            sleep(delay)
            if sample > 0:
                sample *= currentAsymmetricG
            command = {"cur": sample, "pos": pos0}
            sleep(delay)
            data = send_command(command, "current", {}, False, data, device)

        # Step 5
        for sample in samples["currentLine"]:
            command = {"cur": sample, "pos": pos0}
            sleep(delay)
            data = send_command(command, "current", {}, False, data, device)

    # Clean up
    device.close()

    # Plot data
    times = [data["stateTime"], data["stateTime"]]
    desired = [data["desiredCurrent"], data["desiredPosition"]]
    measured = [data["measuredCurrent"], data["measuredPosition"]]
    labels = ["current", "position"]

    for (t, des, meas, label) in zip(times, desired, measured, labels):
        plot(t, des, meas, label)


if __name__ == "__main__":
    main()
